import numpy as np
import scipy.misc
import scipy.ndimage
import scipy.stats
import scipy.io
from PIL import Image

from vmaf.config import VmafConfig
from vmaf.tools.misc import index_and_value_of_min

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


def _gauss_window(lw, sigma):
    sd = float(sigma)
    lw = int(lw)
    weights = [0.0] * (2 * lw + 1)
    weights[lw] = 1.0
    sum = 1.0
    sd *= sd
    for ii in range(1, lw + 1):
        tmp = np.exp(-0.5 * float(ii * ii) / sd)
        weights[lw + ii] = tmp
        weights[lw - ii] = tmp
        sum += 2.0 * tmp
    for ii in range(2 * lw + 1):
        weights[ii] /= sum
    return weights


def _hp_image(image):
    extend_mode = 'reflect'
    image = np.array(image).astype(np.float32)
    w, h = image.shape
    mu_image = np.zeros((w, h))
    _avg_window = _gauss_window(3, 1.0)
    scipy.ndimage.correlate1d(image, _avg_window, 0, mu_image, mode=extend_mode)
    scipy.ndimage.correlate1d(mu_image, _avg_window, 1, mu_image, mode=extend_mode)
    return image - mu_image


def _var_image(hpimg):
    extend_mode = 'reflect'
    w, h = hpimg.shape
    varimg = np.zeros((w, h))
    _var_window = _gauss_window(3, 1.0)
    scipy.ndimage.correlate1d(hpimg**2, _var_window, 0, varimg, mode=extend_mode)
    scipy.ndimage.correlate1d(varimg, _var_window, 1, varimg, mode=extend_mode)
    return varimg


def as_one_hot(label_list):
    return np.eye(2)[np.array(label_list).astype(np.int)]


def create_hp_yuv_4channel(yuvimg):
    yuvimg = yuvimg.astype(np.float32)
    yuvimg /= 255.0
    hp_y = _hp_image(yuvimg[:, :, 0])
    hp_u = _hp_image(yuvimg[:, :, 1])
    hp_v = _hp_image(yuvimg[:, :, 2])
    sigma = np.sqrt(_var_image(hp_y))

    # stack together to make 4 channel image
    return np.dstack((hp_y, hp_u, hp_v, sigma))


def dstack_y_u_v(y, u, v):
    # make y, u, v consistent in size
    if u.shape != y.shape:
        u = np.array(Image.fromarray(u).convert("L").resize(y.shape, Image.BICUBIC))
    if v.shape != y.shape:
        v = np.array(Image.fromarray(v).convert("L").resize(y.shape, Image.BICUBIC))
    return np.dstack((y, u, v))


def midrank(x):
    # [Z J]=sort(x);
    # Z=[Z Z(end)+1];
    # N=length(x);
    # T=zeros(1,N);
    J, Z = zip(*sorted(enumerate(x), key=lambda x:x[1]))
    J = list(J)
    Z = list(Z)
    Z.append(Z[-1]+1)
    N = len(x)
    T = np.zeros(N)

    # i=1;
    # while i<=N
    #     a=i;
    #     j=a;
    #     while Z(j)==Z(a)
    #         j=j+1;
    #     end
    #         b=j-1;
    #     for k=a:b
    #         T(k)=(a+b)/2;
    #     end
    #     i=b+1;
    # end
    i = 1
    while i <= N:
        a = i
        j = a
        while Z[j-1] == Z[a-1]:
            j = j + 1
        b = j - 1
        for k in range(a, b+1):
            T[k-1] = (a + b) / 2
        i = b + 1

    # T(J)=T;
    T2 = np.zeros(N)
    T2[J] = T

    return T2


def calpvalue(aucs, sigma):
    # function pvalue = calpvalue(aucs, sigma)
    # l = [1, -1];
    # z = abs(diff(aucs)) / sqrt(l * sigma * l');
    # pvalue = 2 * (1 - normcdf(z, 0, 1));
    l = np.array([[1, -1]])
    z = np.abs(np.diff(aucs)) / np.sqrt(np.dot(np.dot(l, sigma), l.T))
    # numpy returns a 2-dimensional array with shape (1, 1), extract the element
    z_val = z[0, 0]
    pvalue = 2 * (1 - scipy.stats.norm.cdf(z_val, loc=0, scale=1))
    return pvalue


def _cov_kendall(x):
    """
    x: rows - observation vector 0, 1, 2, ...
    return a covariance matrix based on kendall correlation
    """
    m, n = x.shape
    cov_ = np.zeros([m, m])
    for i in range(m):
        for j in range(i, m):
            # scipy 1.2.0 kendalltau() has an issue with the p-value of two long vectors that are perfectly monotonic
            # see here: https://github.com/scipy/scipy/issues/9611
            # until this is fixed, we bypass the exact calculation method by using the variance approximation (asymptotic method)
            # need a try-except clause: ealier scipy versions do not support a method keywarg
            try:
                kendall, _ = scipy.stats.kendalltau(x[i,:], x[j,:], method='asymptotic')
            except TypeError:
                kendall, _ = scipy.stats.kendalltau(x[i, :], x[j, :])
            cov_[i, j] = kendall
            cov_[j, i] = kendall
    return cov_


def AUC_CI(n_D, n_I, Area):
    # function [CI,SE] = AUC_CI(n_D,n_I,Area)
    #
    # % By Lukas Krasula
    # % Inspired by
    # % *********************  CIAUC  ****************************
    # %   (c) John W Pickering, Novemeber 2009
    # %     Christchurch Kidney Research Group
    # %     University of Otago Christchurch
    # %     New Zealand
    # %
    # %   Last update:  17 July 2012
    # %
    # %	Redistribution and use in source and binary forms, with or without
    # %   modification, are permitted provided that the following conditions are met:
    # %
    # %   * Redistributions of source code must retain the above copyright
    # %     notice, this list of conditions and the following disclaimer.
    # %   * Redistributions in binary form must reproduce the above copyright
    # %     notice, this list of conditions and the following disclaimer in
    # %     the documentation and/or other materials provided with the distribution
    # %
    # % Attribution to John Pickering.
    # % *************************************************************************
    # % n_D - number of different pairs
    # % n_I - number of indifferent pairs
    # % Area - Area under ROC curve

    # Q1=Area/(2-Area);
    # Q2=2*Area*Area/(1+Area);
    Q1 = Area / (2.0 - Area)
    Q2 = 2.0 * Area * Area / (1.0 + Area)

    # SE=sqrt((Area*(1-Area)+(n_D-1)*(Q1-Area*Area)+(n_I-1)*(Q2-Area*Area))/(n_I*n_D));
    SE = np.sqrt((Area * (1.0 - Area) + (n_D-1) * (Q1 - Area * Area) +
                  (n_I - 1.0) * (Q2 - Area*Area)) / (n_I * n_D))

    # CI = 1.96 * SE;
    CI = 1.96 * SE

    return CI, SE


def significanceHM(A, B, AUCs):
    # function [pHM,CI] = significanceHM(A,B,AUCs)
    # % By Lukas Krasula

    assert A.shape[0] == B.shape[0] == AUCs.shape[0]

    # n_met = size(A,1);
    n_met = A.shape[0]

    # CorrA = corr(A','type','Kendall');
    # CorrB = corr(B','type','Kendall');
    CorrA = _cov_kendall(A)
    CorrB = _cov_kendall(B)

    # pHM = ones(n_met);
    # CI = ones(n_met,1);
    # for i=1:n_met-1
    #
    #     [CI(i),SE1] = AUC_CI(size(A,2),size(B,2),AUCs(i));
    #
    #     for j=i+1:n_met
    #         [CI(j),SE2] = AUC_CI(size(A,2),size(B,2),AUCs(j));
    #
    #         load('Hanley_McNeil.mat');
    #
    #         rA = (CorrA(i,j) + CorrB(i,j))/2;
    #         AA = (AUCs(i) + AUCs(j))/2;
    #
    #         [~,rr] = min(abs(rA-rA_vec));
    #         [~,aa] = min(abs(AA-AA_vec));
    #         r = Table_HM(rr,aa);
    #
    #         z = abs(AUCs(i) - AUCs(j)) / sqrt( SE1^2 + SE2^2 + 2*r*SE1*SE2 );
    #         pHM(i,j) = 1-normcdf(z);
    #         pHM(j,i) = pHM(i,j);
    #     end
    # end
    hm_filepath = VmafConfig.tools_resource_path('Hanley_McNeil.mat')
    hm_dict = scipy.io.loadmat(hm_filepath)
    pHM = np.ones([n_met, n_met])
    Table_HM = hm_dict['Table_HM']
    AA_vec = hm_dict['AA_vec']
    rA_vec = hm_dict['rA_vec']
    CI = np.ones(n_met)
    for i in range(1, n_met):
        CI1,SE1 = AUC_CI(A.shape[1], B.shape[1], AUCs[i-1])
        CI[i-1] = CI1

        for j in range(i+1, n_met+1):
            CI2, SE2 = AUC_CI(A.shape[1], B.shape[1], AUCs[j-1])
            CI[j-1] = CI2

            rA = (CorrA[i-1,j-1] + CorrB[i-1,j-1]) / 2
            AA = (AUCs[i-1] + AUCs[j-1]) / 2

            rr, _ = index_and_value_of_min(abs(rA - rA_vec).ravel())
            aa, _ = index_and_value_of_min(abs(AA - AA_vec).ravel())
            r = Table_HM[rr, aa]

            z = abs(AUCs[i - 1] - AUCs[j - 1]) / np.sqrt(SE1 ** 2 + SE2 ** 2 + 2 * r * SE1 * SE2)
            pHM[i-1, j-1] = 1.0 - scipy.stats.norm.cdf(z)
            pHM[j-1, i-1] = pHM[i-1, j-1]

    return pHM, CI


def fastDeLong(samples):
    # %FASTDELONGCOV
    # %The fast version of DeLong's method for computing the covariance of
    # %unadjusted AUC.
    # %% Reference:
    # % @article{sun2014fast,
    # %   title={Fast Implementation of DeLong's Algorithm for Comparing the Areas Under Correlated Receiver Operating Characteristic Curves},
    # %   author={Xu Sun and Weichao Xu},
    # %   journal={IEEE Signal Processing Letters},
    # %   volume={21},
    # %   number={11},
    # %   pages={1389--1393},
    # %   year={2014},
    # %   publisher={IEEE}
    # % }
    # %% [aucs, delongcov] = fastDeLong(samples)
    # %%
    # % Edited by Xu Sun.
    # % Homepage: https://pamixsun.github.io
    # % Version: 2014/12
    # %%

    # if sum(samples.spsizes) ~= size(samples.ratings, 2) || numel(samples.spsizes) ~= 2
    #     error('Argument mismatch error');
    # end
    if np.sum(samples.spsizes) != samples.ratings.shape[1] or len(samples.spsizes) != 2:
        assert False, 'Argument mismatch error'

    # z = samples.ratings;
    # m = samples.spsizes(1);
    # n = samples.spsizes(2);
    # x = z(:, 1 : m);
    # y = z(:, m + 1 : end);
    # k = size(z, 1);
    z = samples.ratings
    m, n = samples.spsizes
    x = z[:, :m]
    y = z[:, m:]
    k = z.shape[0]

    # tx = zeros(k, m);
    # ty = zeros(k, n);
    # tz = zeros(k, m + n);
    # for r = 1 : k
    #     tx(r, :) = midrank(x(r, :));
    #     ty(r, :) = midrank(y(r, :));
    #     tz(r, :) = midrank(z(r, :));
    # end
    tx = np.zeros([k, m])
    ty = np.zeros([k, n])
    tz = np.zeros([k, m + n])
    for r in range(k):
        tx[r, :] = midrank(x[r, :])
        ty[r, :] = midrank(y[r, :])
        tz[r, :] = midrank(z[r, :])

    # % tz
    # aucs = sum(tz(:, 1 : m), 2) / m / n - (m + 1) / 2 / n;
    # v01 = (tz(:, 1 : m) - tx(:, :)) / n;
    # v10 = 1 - (tz(:, m + 1 : end) - ty(:, :)) / m;
    # sx = cov(v01')';
    # sy = cov(v10')';
    # delongcov = sx / m + sy / n;
    aucs = np.sum(tz[:, :m], axis=1) / m / n - float(m + 1.0) / 2.0 / n
    v01 = (tz[:, :m] - tx[:, :]) / n
    v10 = 1.0 - (tz[:, m:] - ty[:, :]) / m
    sx = np.cov(v01)
    sy = np.cov(v10)
    delongcov = sx / m + sy / n

    return aucs, delongcov, v01, v10


def significanceBinomial(p1, p2, N):
    # function pValue = significanceBinomial(p1,p2,N)
    # p = (p1+p2) / 2;
    # sigmaP1P2 = sqrt(p*(1-p)*2/N);
    # z = abs(p1-p2)/sigmaP1P2;
    # pValue = 2*(1 - normcdf(z, 0, 1));
    p = (p1 + p2) / 2.0
    sigmaP1P2 = np.sqrt(p * (1.0 - p) * 2.0 / N)
    z = abs(p1 - p2) / sigmaP1P2
    pValue = 2.0 * (1.0 - scipy.stats.norm.cdf(z, 0.0, 1.0))

    return pValue



import os
import pickle
import sys

import numpy as np
import scipy.stats
from sklearn.metrics import f1_score

from vmaf.tools.decorator import override

try:
    import tensorflow.compat.v1 as tf
    tf.disable_v2_behavior()
except ImportError:
    tf = None

from vmaf.core.h5py_mixin import H5pyMixin
from vmaf.core.train_test_model import RawVideoTrainTestModelMixin, TrainTestModel, \
    ClassifierMixin
from vmaf.tools.sigproc import as_one_hot, create_hp_yuv_4channel, dstack_y_u_v
from vmaf.tools.misc import get_dir_without_last_slash

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class NeuralNetTrainTestModel(RawVideoTrainTestModelMixin,
                              TrainTestModel,
                              # order affects whose _assert_dimension
                              # gets called
                              H5pyMixin):

    DEFAULT_N_EPOCHS = 30
    DEFAULT_LEARNING_RATE = 1e-3
    DEFAULT_PATCH_WIDTH = 50
    DEFAULT_PATCH_HEIGHT = 50
    DEFAULT_PATCHES_PER_FRAME = 1
    DEFAULT_BATCH_SIZE = 20
    DEFAULT_SEED = None

    @property
    def n_epochs(self):
        if self.param_dict is not None and 'n_epochs' in self.param_dict:
            return self.param_dict['n_epochs']
        else:
            return self.DEFAULT_N_EPOCHS

    @property
    def learning_rate(self):
        if self.param_dict is not None and 'learning_rate' in self.param_dict:
            return self.param_dict['learning_rate']
        else:
            return self.DEFAULT_LEARNING_RATE

    @property
    def patch_width(self):
        if self.param_dict is not None and 'patch_width' in self.param_dict:
            return self.param_dict['patch_width']
        else:
            return self.DEFAULT_PATCH_WIDTH

    @property
    def patch_height(self):
        if self.param_dict is not None and 'patch_height' in self.param_dict:
            return self.param_dict['patch_height']
        else:
            return self.DEFAULT_PATCH_HEIGHT

    @property
    def patches_per_frame(self):
        if self.param_dict is not None and \
           'patches_per_frame' in self.param_dict:
            return self.param_dict['patches_per_frame']
        else:
            return self.DEFAULT_PATCHES_PER_FRAME

    @property
    def batch_size(self):
        if self.param_dict is not None and 'batch_size' in self.param_dict:
            return self.param_dict['batch_size']
        else:
            return self.DEFAULT_BATCH_SIZE

    @property
    def seed(self):
        if self.param_dict is not None and 'seed' in self.param_dict:
            return self.param_dict['seed']
        else:
            return self.DEFAULT_SEED

    @property
    def checkpoints_dir(self):
        # if None, won't output checkpoints
        if self.optional_dict2 is not None and \
           'checkpoints_dir' in self.optional_dict2:
            return self.optional_dict2['checkpoints_dir']
        else:
            return None

    def _assert_args(self):
        super(NeuralNetTrainTestModel, self)._assert_args()
        self.assert_h5py_file()  # assert h5py_file in self.optional_dict2

    @staticmethod
    def _assert_xs(xs):
        # for now, force input xys or xs having 'dis_y', 'dis_u' and 'dis_v'
        assert 'dis_y' in xs
        assert 'dis_u' in xs
        assert 'dis_v' in xs

    @override(TrainTestModel)
    def train(self, xys):
        self._assert_xs(xys)

        self.model_type = self.TYPE

        assert 'label' in xys
        assert 'content_id' in xys

        # this makes sure the order of features are normalized, and each
        # dimension of xys_2d is consistent with feature_names
        feature_names = sorted(xys.keys())
        feature_names.remove('label')
        feature_names.remove('content_id')
        self.feature_names = feature_names

        self.norm_type = 'none'  # no conventional data normalization

        patches_cache, labels_cache = self._populate_patches_and_labels(
            feature_names, xys)

        model = self._train(patches_cache, labels_cache)
        self.model = model

    @override(TrainTestModel)
    def predict(self, xs):

        self._assert_xs(xs)
        self._assert_trained()
        for name in self.feature_names:
            assert name in xs

        feature_names = self.feature_names

        # loop through xs
        len_xs = len(xs[feature_names[0]])
        ys_label_pred = []
        for i in range(len_xs):

            sys.stdout.write("Evaluating test data: %d / %d\r" % (i, len_xs))
            sys.stdout.flush()

            # create single x
            x = {}
            for feature_name in feature_names:
                x[feature_name] = [xs[feature_name][i]]

            # extract patches
            patches_cache = self._populate_patches(feature_names, x)

            # predict
            y_label_pred = self._predict(self.model, patches_cache)
            ys_label_pred.append(y_label_pred)

        return {'ys_label_pred': ys_label_pred}

    def _create_patch_and_label_dataset(self, total_frames, overwrite=True):

        patches_dims = (total_frames * self.patches_per_frame,
                        self.patch_height, self.patch_width, self.n_channels)
        if overwrite:
            try:
                del self.h5py_file['patches']
            except KeyError:
                pass
        patches_cache = self.h5py_file.create_dataset('patches', patches_dims, dtype='float')
        patches_cache.dims[0].label = 'batch'
        patches_cache.dims[1].label = 'height'
        patches_cache.dims[2].label = 'width'
        patches_cache.dims[3].label = 'channel'

        labels_dims = (total_frames * self.patches_per_frame, )
        if overwrite:
            try:
                del self.h5py_file['labels']
            except KeyError:
                pass
        labels_cache = self.h5py_file.create_dataset('labels', labels_dims, dtype='uint8')

        return patches_cache, labels_cache

    def _populate_patches_and_labels(self, xkeys, xys, mode='train'):

        np.random.seed(self.seed)

        # estimate the size and create h5py dataset
        total_frames = self._get_total_frames(xys)

        patches_cache, labels_cache = self._create_patch_and_label_dataset(
            total_frames)

        assert 'dis_y' in xkeys
        assert 'dis_u' in xkeys
        assert 'dis_v' in xkeys

        yss = xys['dis_y']  # yss: Y * frames * videos
        uss = xys['dis_u']
        vss = xys['dis_v']
        if mode == 'train':
            assert 'label' in xys
            labels = xys['label']
        elif mode == 'test':
            labels = [None for _ in range(len(yss))]
        else:
            assert False

        assert len(yss) == len(uss) == len(vss) == len(labels)

        patch_idx = 0
        for ys, us, vs, label in zip(yss, uss, vss, labels):  # iterate videos
            assert len(ys) == len(us) == len(vs)
            for y, u, v in zip(ys, us, vs):  # iterate frames

                yuvimg = dstack_y_u_v(y, u, v)

                img = create_hp_yuv_4channel(yuvimg)

                h, w, c = img.shape

                adj_h = h - self.patch_height
                adj_w = w - self.patch_width

                iv, jv = np.meshgrid(np.arange(adj_h), np.arange(adj_w), sparse=False, indexing='ij')
                iv = iv.reshape(-1)
                jv = jv.reshape(-1)

                idx = np.random.permutation(adj_h * adj_w)

                iv = iv[idx]
                jv = jv[idx]

                patches_found = 0
                for (yy, xx) in zip(iv, jv):
                    patches_cache[patch_idx] = img[yy: yy + self.patch_height,
                                                   xx: xx + self.patch_width]
                    if mode == 'train':
                        labels_cache[patch_idx] = label

                    patches_found += 1
                    patch_idx += 1
                    if patches_found >= self.patches_per_frame:
                        break

        return patches_cache, labels_cache

    def _populate_patches(self, xkeys, xs):
        # reuse _populate_patches_and_labels to do the job
        patches_cache, _ = self._populate_patches_and_labels(
            xkeys, xys=xs, mode='test')
        return patches_cache

    def _get_total_frames(self, xys):
        yss = xys['dis_y']  # yss
        return np.sum(list(map(lambda ys: len(ys), yss)))

    @override(TrainTestModel)
    def to_file(self, filename, **more):

        self._assert_trained()

        # special handling for tensorflow: save .model differently
        model_dict_copy = self.model_dict.copy()
        model_dict_copy['model'] = None
        info_to_save = {'param_dict': self.param_dict,
                        'model_dict': model_dict_copy}

        saver = tf.train.Saver()
        sess = self.model['sess']
        saver.save(sess, filename + '.model')

        with open(filename, 'wb') as file:
            pickle.dump(info_to_save, file)

    @classmethod
    @override(TrainTestModel)
    def from_file(cls, filename, logger=None, optional_dict2=None, **more):
        fmt = more['format'] if 'format' in more else 'pkl'
        assert fmt in ['pkl'], f'format must be pkl but got {fmt}'

        assert os.path.exists(filename), 'File name {} does not exist.'.format(filename)
        with open(filename, 'rb') as file:
            info_loaded = pickle.load(file)

        train_test_model = cls(param_dict={}, logger=logger,
                               optional_dict2=optional_dict2)
        train_test_model.param_dict = info_loaded['param_dict']
        train_test_model.model_dict = info_loaded['model_dict']

        # == special handling of tensorflow: load .model differently ==

        input_image_batch, logits, y_, y_p, W_conv0, W_conv1, loss, train_step \
            = cls.create_tf_variables(train_test_model.param_dict)

        saver = tf.train.Saver()
        sess = tf.Session()
        checkpoint_dir = get_dir_without_last_slash(filename)
        ckpt = tf.train.get_checkpoint_state(checkpoint_dir)
        if ckpt and ckpt.model_checkpoint_path:
            saver.restore(sess, ckpt.model_checkpoint_path)
        else:
            assert False
        model = {
            'sess': sess,
            'y_p': y_p,
            'input_image_batch': input_image_batch,
        }
        train_test_model.model_dict['model'] = model

        return train_test_model

    @staticmethod
    @override(TrainTestModel)
    def delete(filename, **more):
        fmt = more['format'] if 'format' in more else 'pkl'
        assert fmt in ['pkl'], f'format must be pkl but got {fmt}'

        if os.path.exists(filename):
            os.remove(filename)
        if os.path.exists(filename + '.model'):
            os.remove(filename + '.model')
        if os.path.exists(filename + '.model.meta'):
            os.remove(filename + '.model.meta')
        filedir = get_dir_without_last_slash(filename)
        if os.path.exists(filedir + '/checkpoint'):
            os.remove(filedir + '/checkpoint')

    @classmethod
    def reset(cls):
        super(NeuralNetTrainTestModel, cls).reset()

        # reset tensorflow to avoid any memory
        tf.reset_default_graph()


class ToddNoiseClassifierTrainTestModel(NeuralNetTrainTestModel, ClassifierMixin):

    TYPE = "TODDNOISECLASSIFIER"
    VERSION = "0.1"

    # override NeuralNetTrainTestModel.DEFAULT_PATCHES_PER_FRAME
    DEFAULT_PATCHES_PER_FRAME = 10

    n_channels = 4
    n_filters = 5
    fsize0 = 5
    fsize1 = 3

    def _train(self, patches, labels):

        assert len(patches) == len(labels)

        # randomly split data into training and validation set
        num_data = len(patches)
        indices = np.random.permutation(num_data)
        num_train_data = int(num_data / 2)  # do even split
        train_indices = indices[:num_train_data]
        validate_indices = indices[num_train_data:]
        train_posindices = list(filter(lambda idx: labels[idx] == 1, train_indices))
        train_negindices = list(filter(lambda idx: labels[idx] == 0, train_indices))

        input_image_batch, logits, y_, y_p, W_conv0, W_conv1, loss, train_step \
            = self.create_tf_variables(self.param_dict)

        init = tf.initialize_all_variables()
        sess = tf.Session()
        sess.run(init)

        saver = None

        # loop through and compute training/testing loss per epoch
        f1score_per_epoch = []
        loss_per_epoch = []
        for j in range(self.n_epochs):
            print("")

            print("******************** EPOCH %d / %d ********************" % (j, self.n_epochs))

            # train
            train_loss, train_score = self._evaluate_on_patches(
                patches, labels, input_image_batch, loss,
                sess, train_indices, y_, y_p, "train")

            print("")

            # validate
            validate_loss, validate_score = self._evaluate_on_patches(
                patches, labels, input_image_batch, loss,
                sess, validate_indices, y_, y_p, "validate")

            print("")

            print("f1 train %g, f1 validate %g, loss train %g, loss validate %g"
                  % (train_score, validate_score, train_loss, validate_loss))
            f1score_per_epoch.append([train_score, validate_score])
            loss_per_epoch.append([train_loss, validate_loss])

            if self.checkpoints_dir:
                if saver is None:
                    saver = tf.train.Saver(max_to_keep=0)
                outputfile = "%s/model_epoch_%d.ckpt" % (self.checkpoints_dir, j,)
                print("Checkpointing -> %s" % (outputfile,))
                saver.save(sess, outputfile)

            halfbatch = self.batch_size // 2

            # here, we enforce balanced training, so that if we would like to
            # use hinge loss, we will always have representatives from both
            # target classes

            n_iterations = np.min((len(train_posindices) // halfbatch,
                                   len(train_negindices) // halfbatch))

            for i in range(n_iterations):
                # must sort, since h5py needs ordered indices
                poslst = np.sort(train_posindices[i * halfbatch: (i+1) * halfbatch]).tolist()
                neglst = np.sort(train_negindices[i * halfbatch: (i+1) * halfbatch]).tolist()

                X_batch = np.vstack((
                    patches[poslst],
                    patches[neglst],
                ))
                y_batch = np.vstack((
                    as_one_hot(labels[poslst]),
                    as_one_hot(labels[neglst])
                ))

                sys.stdout.write("Training: %d / %d\r" % (i, n_iterations))
                sys.stdout.flush()

                sess.run(train_step, feed_dict={input_image_batch: X_batch, y_: y_batch})

            np.random.shuffle(train_posindices)
            np.random.shuffle(train_negindices)

            print("")

        model = {
            'sess': sess,
            'y_p': y_p,
            'input_image_batch': input_image_batch,
        }

        return model

    def _evaluate_on_patches(self, patches_in, labels_in, input_image_batch,
                             loss_in, sess, indices, y_, y_p, type_="train"):
        ys_pred = []
        ys_true = []
        loss_cum = 0.0
        n_steps = len(indices) // self.batch_size
        for M in range(n_steps):
            sys.stdout.write("Evaluating {type} data: {M} / {nstep}\r"
                             .format(type=type_, M=M, nstep=n_steps))
            sys.stdout.flush()
            curr_indices = indices[M * self.batch_size: (M + 1) * self.batch_size].tolist()
            patches = list(map(lambda idx: patches_in[idx], curr_indices))
            labels_ = list(map(lambda idx: labels_in[idx], curr_indices))
            labels = as_one_hot(labels_)
            y_pred, loss = sess.run([y_p, loss_in], feed_dict={input_image_batch: patches, y_: labels})
            ys_pred = np.hstack((ys_pred, y_pred))
            ys_true = np.hstack((ys_true, labels_))
            loss_cum += loss
        loss_cum /= n_steps
        score = f1_score(ys_true, ys_pred)
        return loss_cum, score

    @classmethod
    def _predict(cls, model, patches):

        sess = model['sess']
        y_p = model['y_p']
        input_image_batch = model['input_image_batch']

        ys_patche_pred = sess.run(y_p, feed_dict={input_image_batch: patches})

        # predict by majority voting on all patches
        y_pred = scipy.stats.mode(ys_patche_pred)[0][0]

        return y_pred

    @classmethod
    def create_tf_variables(cls, param_dict):
        def weight_variable(shape, name="test", seed=None):
            return tf.Variable(tf.random_normal(shape, stddev=0.1, seed=seed), name=name)

        def conv2d(x, W):
            return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding='VALID')

        if param_dict is not None and 'patch_width' in param_dict:
            patch_width = param_dict['patch_width']
        else:
            patch_width = cls.DEFAULT_PATCH_WIDTH

        if param_dict is not None and 'patch_height' in param_dict:
            patch_height = param_dict['patch_height']
        else:
            patch_height = cls.DEFAULT_PATCH_HEIGHT

        if param_dict is not None and 'learning_rate' in param_dict:
            learning_rate = param_dict['learning_rate']
        else:
            learning_rate = cls.DEFAULT_LEARNING_RATE

        if param_dict is not None and 'seed' in param_dict:
            seed = param_dict['seed']
        else:
            seed = cls.DEFAULT_SEED

        tf.set_random_seed(seed)

        n_channels = cls.n_channels
        response_map_size_width = patch_width - (cls.fsize0 // 2) * 2
        response_map_size_width -= (cls.fsize1 // 2) * 2
        response_map_size_height = patch_height - (cls.fsize0 // 2) * 2
        response_map_size_height -= (cls.fsize1 // 2) * 2
        response_map_size = response_map_size_height * response_map_size_width
        input_image_batch = tf.placeholder(
            tf.float32, shape=[None, patch_height, patch_width, n_channels])
        y_ = tf.placeholder(tf.float32, shape=[None, 2])
        W_conv0 = weight_variable(
            [cls.fsize0, cls.fsize0, n_channels, cls.n_filters], "W_conv0")
        W_conv1 = weight_variable(
            [cls.fsize1, cls.fsize1, cls.n_filters, 2], "W_conv1")

        # convolutional layer 1
        h_conv0 = conv2d(input_image_batch, W_conv0)
        h_conv0_elu = tf.nn.elu(h_conv0)

        # layer 2, which is just an output layer
        h_conv1 = conv2d(h_conv0_elu, W_conv1)
        h_conv1_elu = tf.nn.elu(h_conv1)
        h_conv1_elu_flat = tf.reshape(h_conv1_elu, [-1, response_map_size, 2])

        logits = tf.reduce_mean(h_conv1_elu_flat, 1)
        logits_norm = tf.nn.softmax(logits)

        y_p = tf.argmax(logits_norm, 1)
        loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=logits, labels=y_))
        train_step = tf.train.AdamOptimizer(learning_rate).minimize(loss)

        return input_image_batch, logits, y_, y_p, W_conv0, W_conv1, loss, train_step



"""scanf.py: scanf-style input for Python.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

The initial motivation for this module was based on a posting on
Python-tutor:

    http://mail.python.org/pipermail/tutor/2004-July/030480.html

I haven't been able to find a nice module to do scanf-style input.
Even the Library Reference recommends regular expressions as a
substitute:

    http://docs.python.org/lib/node109.html

But there appears to have been activity about this on python-list:

    http://aspn.activestate.com/ASPN/Mail/Message/python-list/785450


Still, let's see if we can get a close equivalent scanf() in place.
At the least, it'll be fun for me, and it might be useful for people
who are still recovering from C.  *grin*


Functions provided:

    scanf(formatString) -- formatted scanning across stdin

    sscanf(sourceString, formatString) -- formated scanning across strings

    fscanf(sourceFile, formatString) -- formated scanning across files


The behavior of this scanf() will be slightly different from that
defined in C, because, in truth, I'm a little lazy, and am not quite
sure if people will need all of scanf's features in typical Python
programming.


But let's first show what conversions this scanf() will support.
Format strings are of the following form:

    % [*] [width] [format]

where [*] and [width] are optional, and [format] is mandatory.  The
optional flags modify the format.

    *         suppresses variable capture.
    width     maximum character width.


We support the following scanf conversion formats (copied from K&R):

    d    decimal integer.

    i    integer.  The integer may be in octal (leading zero) or
         hexadecimal (leading 0x or 0X).  ## fixme

    o    octal integer (with or without leading zero).  ## fixme

    x    hexadecimal integer (with or without leading 0x or 0X)   ## fixme

    c    characters.  The next input characters (default 1) are
         placed at the indicated spot.  The normal skip over white space
         is suppressed; to read the next non-white space character, use
         %1s.

    s    character string (not quoted).

    f    floating-point number with optional sign and optional decimal point.

    %    literal %; no assignment is made.


Literal characters can appear in the scanf format string: they must
match the same characters in the input.

There is no guarantee of what happens if calls to scanf are mixed with
other input functions.  See the BUGS section below for details on this.


If the input doesn't conform to the format string, a FormatError is
raised.


Example format strings:

    "%d %d"         Two decimal integers.

    "%d.%d.%d.%d"   Four decimal integers, separated by literal periods.
                    The periods won't be captured.

    "hello %s"      Literally matches "hello" followed by any number of
                    spaces, followed by a captured word.


There's also an interface for calling the internal function bscanf()
that works on CharacterBuffer types, if in the future there is
something that supports getc() and ungetc() natively.  There's also an
undocumented compile() function that takes format strings and returns
a function that can scan through CharacterBuffers.  Ooops, I guess I
just documented it.  *grin*


######################################################################


BUGS and GOTCHAS:

One major problem that I'm running into is a lack of ungetc(); it
would be nice if there were such a function in Python, but I can't
find it.  I have to simulate it by using a CharacterBuffer object, but
it's not an ideal solution.

So at most, you may lose a single character to the internal buffers
maintained by this module if you use scanf().  The other two *scanf()
functions, thankfully, aren't effected by this problem, since I can
simulate ungetc() more accurately by using seek() in the other two
cases.

If you really need to get that buffered character back, you can grab
it through _STDIN.lastChar, though manually fiddling with this is not
recommended.

So use scanf() with the following caveat: unlike C's stdin(), this
version scanf() can't be interchanged with calls to other input
functions without some kind of weird side effect.  We keep a
one-character buffer into stdin, so at most you might lose one
character to the internal buffers.

fscanf() is only allowed to work on things that support both read(1)
and seek(1, -1), since then I can reliably do a ungetch-like thing.

scanf("%s") can be dangerous in a hostile environment, since it's very
possible for something to pass in a huge string without spaces.  So use
an explicit width instead if you can help it.

######################################################################

LICENSE

Copyright (c) 2008, Danny yoo
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import string
import sys
import unittest

from io import StringIO


__all__ = ['scanf', 'sscanf', 'fscanf']
__version__ = '1.0'

# We keep a few sets as module variables just to incur the cost of constructing them just once.
_DIGIT_SET = set(string.digits)
_HEX_SET = set("0123456789ABCDEFabcdef")
_OCT_SET = set("01234567")
_PLUS_MINUS_SET = set("+-")
_WHITESPACE_SET = set(string.whitespace)


class CharacterBuffer(object):
    """A CharacterBuffer allows us to get a character, and to "unget" a
    character.  Abstract class"""

    def getch(self):
        """Returns the next character.  If there are no more characters
        left in the stream, returns the empty string."""
        pass # implement me!

    def ungetch(self, ch):
        """Tries to put back a character.  Can be called at most once
        between calls to getch()."""
        pass # implement me!

    def scanCharacterSet(self, characterSet, maxChars=0):
        """Support function that scans across a buffer till we hit
        something outside the allowable characterSet."""
        return self.scanPredicate(lambda ch: ch in characterSet, maxChars)

    def scanPredicate(self, predicate, maxChars=0):
        """Support function that scans across a buffer till we hit
        something outside what's allowable by the predicate."""
        chars = []
        countChars = 0
        while True:
            if maxChars != 0 and countChars >= maxChars:
                break
            ch = self.getch()
            if ch != '' and predicate(ch):
                chars.append(ch)
                countChars += 1
            else:
                self.ungetch(ch)
                break
        return ''.join(chars)


class CharacterBufferFromIterable(CharacterBuffer):
    """Implementation of CharacterBuffers for iterable things.
    We keep a 'lastChar' attribute to simulate ungetc()."""
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.lastChar = ''

    def getch(self):
        if self.lastChar == '':
            try:
                return next(self.iterator)
            except StopIteration:
                return ''
        else:
            (ch, self.lastChar) = (self.lastChar, '')
            return ch

    def ungetch(self, ch):
        self.lastChar = ch


class CharacterBufferFromStdin(CharacterBuffer):
    def __init__(self, stdin):
        self.stdin = stdin
        self.lastChar = ''

    def getch(self):
        if self.lastChar == '':
            try:
                return self.stdin.read(1)
            except StopIteration:
                return ''
        else:
            (ch, self.lastChar) = (self.lastChar, '')
            return ch

    def ungetch(self, ch):
        self.lastChar = ch


class CharacterBufferFromFile(CharacterBuffer):
    """Implementation of CharacterBuffers for files.  We use the native
    read(1) and seek() calls, so we don't have to do so much magic."""
    def __init__(self, myfile):
        self.myfile = myfile

    def getch(self):
        return self.myfile.read(1)

    def ungetch(self, ch):
        self.myfile.seek(- len(ch), 1)


def readiter(inputFile, *args):
    """Returns an iterator that calls read(*args) on the inputFile."""
    while True:
        ch = inputFile.read(*args)
        if ch:
            yield ch
        else:
            raise StopIteration


def isIterable(thing):
    """Returns true if 'thing' looks iterable."""
    try:
        iter(thing)
    except TypeError:
        return False
    return True


def isFileLike(thing):
    """Returns true if thing looks like a file."""
    if hasattr(thing, "read") and hasattr(thing, "seek"):
        try:
            thing.seek(1, 1)
            thing.seek(-1, 1)
            return True
        except IOError:
            pass
    return False


def makeCharBuffer(thing):
    """Try to coerse 'thing' into a CharacterBuffer.  'thing' can be
    an instance of:

        1.  CharacterBuffer
        2.  A file-like object,
        3.  An iterable.

    makeCharBuffer() will make guesses in that order.
    """
    if isinstance(thing, CharacterBuffer):
        return thing
    elif isFileLike(thing):
        # this check must come before isIterable, since files
        # provide a line-based iterator that we don't want to use.
        # Plus we want to take advantage of file.seek()
        return CharacterBufferFromFile(thing)
    elif isIterable(thing):
        return CharacterBufferFromIterable(thing)
    else:
        raise ValueError("Can't coerse %r to CharacterBuffer" % thing)


class CappedBuffer(CharacterBuffer):
    """Implementation of a buffer that caps the number of bytes we can
    getch().  The cap may or may not include whitespace characters."""
    def __init__(self, buffer, width, ignoreWhitespace=False):
        self.buffer = buffer
        self.bytesRead = 0
        self.width = width
        self.ignoreWhitespace = ignoreWhitespace

    def getch(self):
        if self.bytesRead < self.width:
            nextChar = self.buffer.getch()
            if not self.isIgnoredChar(nextChar):
                self.bytesRead += len(nextChar)
            return nextChar
        else:
            return ''

    def isIgnoredChar(self, ch):
        return self.ignoreWhitespace and isWhitespaceChar(ch)

    def ungetch(self, ch):
        self.buffer.ungetch(ch)
        if not self.isIgnoredChar(ch):
            self.bytesRead -= len(ch)
        # make sure wacky things don't happen when ungetch()ing.
        assert self.bytesRead >= 0


class FormatError(ValueError):
    """A FormatError is raised if we run into errors while scanning
    for input."""
    pass

class IncompleteCaptureError(ValueError):
    """The *scanf() functions raise IncompleteCaptureError if a problem
    occurs doing scanning."""
    pass


"""We keep a module-level STDIN CharacterBuffer, so that we can call
scanf() several times and not lose characters between invocations."""
_STDIN = CharacterBufferFromStdin(sys.stdin)


def scanf(formatString):
    """scanf(formatString) -> tuple

Scans standard input for formats specified in the formatString.  See
module's docs for list of supported format characters."""
    return bscanf(_STDIN, formatString)


def sscanf(inputString, formatString):
    """sscanf(inputString, formatString) -> tuple

Scans inputString for formats specified in the formatString.  See
module's docs for list of supported format characters."""
    return bscanf(CharacterBufferFromIterable(inputString), formatString)


def fscanf(inputFile, formatString):
    """fscanf(inputFile, formatString) -> tuple

Scans inputFile for formats specified in the formatString.  See
module's docs for list of supported format characters."""
    buffer = CharacterBufferFromFile(inputFile)
    return bscanf(buffer, formatString)


def bscanf(buffer, formatString):
    """fscanf(buffer, formatString) -> tuple

Scans a CharacterBuffer 'buffer' for formats specified in the
formatString.  See scanf module's docs for list of supported format
characters."""
    # TODO: we may want to do some caching here of compiled formatStrings,
    # similar to that of the 're' module.
    parser = compile(formatString)
    return parser(buffer)


def isWhitespaceChar(ch, _set=_WHITESPACE_SET):
    """Returns true if the charcter looks like whitespace.
    We follow the definition of C's isspace() function.
    """
    return ch in _set


def handleWhitespace(buffer):
    """Scans for whitespace.  Returns all the whitespace it collects."""
    chars = []
    while True:
        ch = buffer.getch()
        if isWhitespaceChar(ch):
            chars.append(ch)
        else:
            buffer.ungetch(ch)
            break
    return ''.join(chars)


def handleDecimalInt(buffer, optional=False, allowLeadingWhitespace=True):
    """Tries to scan for an integer.  If 'optional' is set to False,
    returns None if an integer can't be successfully scanned."""
    if allowLeadingWhitespace:
        handleWhitespace(buffer)  # eat leading spaces
    chars = []
    chars += buffer.scanCharacterSet(_PLUS_MINUS_SET, 1)
    chars += buffer.scanCharacterSet(_DIGIT_SET)
    try:
        return int(''.join(chars), 10)
    except ValueError:
        if optional:
            return None
        raise FormatError("invalid literal characters: %s" % ''.join(chars))


def handleOct(buffer):
    chars = []
    chars += buffer.scanCharacterSet(_PLUS_MINUS_SET)
    chars += buffer.scanCharacterSet(_OCT_SET)
    try:
        return int(''.join(chars), 8)
    except ValueError:
        raise FormatError("invalid literal characters: %s" % ''.join(chars))


def handleInt(buffer, base=0):
    chars = []
    chars += buffer.scanCharacterSet(_PLUS_MINUS_SET)
    chars += buffer.scanCharacterSet("0")
    if chars and chars[-1] == '0':
        chars += buffer.scanCharacterSet("xX")
    chars += buffer.scanCharacterSet(_HEX_SET)
    try:
        return int(''.join(chars), base)
    except ValueError:
        raise FormatError("invalid literal characters: %s" % ''.join(chars))


def handleHex(buffer):
    return handleInt(buffer, 16)


def handleFloat(buffer, allowLeadingWhitespace=True):
    if allowLeadingWhitespace:
        handleWhitespace(buffer) # eat leading whitespace
    chars = []
    chars += buffer.scanCharacterSet(_PLUS_MINUS_SET)
    chars += buffer.scanCharacterSet(_DIGIT_SET)
    chars += buffer.scanCharacterSet(".")
    chars += buffer.scanCharacterSet(_DIGIT_SET)
    chars += buffer.scanCharacterSet("eE")
    chars += buffer.scanCharacterSet(_PLUS_MINUS_SET)
    chars += buffer.scanCharacterSet(_DIGIT_SET)
    try:
        return float(''.join(chars))
    except ValueError:
        raise FormatError("invalid literal characters: %s" % ''.join(chars))


def handleChars(buffer,
                allowLeadingWhitespace=False,
                isBadCharacter=lambda ch: False,
                optional=False):
    """Read as many characters are there are in the buffer."""
    if allowLeadingWhitespace:
        handleWhitespace(buffer)
    chars = []
    chars += buffer.scanPredicate(lambda ch: not isBadCharacter(ch))
    if chars:
        return ''.join(chars)
    else:
        if optional:
            return None
        raise FormatError("Empty buffer.")


def handleString(buffer, allowLeadingWhitespace=True):
    """Reading a string format is just an application of reading
    characters (skipping leading spaces, and reading up to space)."""
    return handleChars(buffer,
                       allowLeadingWhitespace=allowLeadingWhitespace,
                       isBadCharacter=isWhitespaceChar)


def makeHandleLiteral(literal):
    def f(buffer, optional=False):
        ch = buffer.getch()
        if ch == literal:
            return ch
        else:
            buffer.ungetch(ch)
            if optional:
                return None
            raise FormatError("%s != %s" % (literal, ch))
    return f


def makeWidthLimitedHandler(handler, width, ignoreWhitespace=False):
    """Constructs a Handler that caps the number of bytes that can be read
    from the byte buffer."""
    def f(buffer):
        return handler(CappedBuffer(buffer, width, ignoreWhitespace))
    return f


"""Just for kicks: handleChar is a handler for a single character."""
handleChar = makeWidthLimitedHandler(handleChars, 1, ignoreWhitespace=False)


def makeIgnoredHandler(handler):
    def f(buffer):
        handler(buffer)
        return None
    return f


class CompiledPattern:
    def __init__(self, handlers, formatString):
        self.handlers = handlers
        self.formatString = formatString

    def __call__(self, buffer):
        results = []
        try:
            for h in self.handlers:
                value = h(buffer)
                # We use None as the sentinel value that ignored handlers
                # will emit.
                if value is not None:
                    results.append(value)
            return tuple(results)
        except FormatError as e:
            raise IncompleteCaptureError(e, tuple(results))

    def __repr__(self):
        return "compile(%r)" % self.formatString


def compile(formatString):
    """Given a format string, emits a new CompiledPattern that eats
    CharacterBuffers and returns captured values as a tuple.

    If there's a failure during scanning, raises IncompleteCaptureError,
    with args being a two-tuple of the FormatError, and the results that
    were captured before the error occurred.
    """
    handlers = []
    formatBuffer = CharacterBufferFromIterable(formatString)
    while True:
        ch = formatBuffer.getch()
        if ch == '':
            break
        if isWhitespaceChar(ch):
            handleWhitespace(formatBuffer)
            handlers.append(makeIgnoredHandler(handleWhitespace))
        elif ch == '%':
            handlers.append(_compileFormat(formatBuffer))
        else:
            handlers.append(makeIgnoredHandler(makeHandleLiteral(ch)))
    return CompiledPattern(handlers, formatString)


def _compileFormat(formatBuffer):
    def readOptionalSuppression():
        f = makeHandleLiteral("*")
        return f(formatBuffer, optional=True) == "*"

    def readOptionalWidth():
        return handleDecimalInt(formatBuffer,
                                optional=True,
                                allowLeadingWhitespace=False)

    def readFormat():
        return formatBuffer.getch()  # Finally, read the format

    suppression = readOptionalSuppression()
    width = readOptionalWidth()
    formatCh = readFormat()
    handler = makeFormattedHandler(suppression, width, formatCh)
    if handler:
        return handler
    else:
        # At this point, since we couldn't figure out the format, die loudly.
        raise FormatError("Invalid format character %s" % formatCh)


_FORMAT_HANDLERS = {'d': handleDecimalInt,
                    'i': handleInt,
                    'x': handleHex,
                    'o': handleOct,
                    's': handleString,
                    'f': handleFloat,
                    '%': makeIgnoredHandler(makeHandleLiteral('%'))
                    }


def makeFormattedHandler(suppression, width, formatCh):
    """Given suppression, width, and a formatType, returns a function
    that eats a buffer and returns that thing."""
    def applySuppression(handler):
        if suppression:
            return makeIgnoredHandler(handler)
        return handler

    def applyWidth(handler):
        if width is None:
            return makeWidthLimitedHandler(handler, width,
                                           ignoreWhitespace=True)
        return handler

    # 'c' is a special case: it's the only handler that can't ignore
    # whitespace.
    if formatCh == 'c':
        if width is None:
            return applySuppression(handleChar)
        else:
            return applySuppression(
                makeWidthLimitedHandler(handleChars, width,
                                        ignoreWhitespace=False))
    if formatCh in _FORMAT_HANDLERS:
        return applySuppression(applyWidth(_FORMAT_HANDLERS[formatCh]))
    else:
        return None


######################################################################
#
# Of course we have test cases.  *grin*
#

class ScanfTests(unittest.TestCase):
    def bufferFromString(self, s):
        return CharacterBufferFromIterable(s)

    def testBufferFromString(self):
        b = self.bufferFromString("hello")
        for letter in list('hello'):
            self.assertEqual(letter, b.getch())
        self.assertEqual('', b.getch())

    def testCharacterSetScanning(self):
        b = makeCharBuffer("+++-+++++1234")
        self.assertEqual("+++", b.scanCharacterSet(set("+")))
        self.assertEqual("", b.scanCharacterSet(set("+")))
        self.assertEqual("-", b.scanCharacterSet(set("-")))
        self.assertEqual("+", b.scanCharacterSet(set("+"), 1))

    def testPredicateScanning(self):
        b = makeCharBuffer("+++-+++++1234")
        self.assertEqual("+++", b.scanPredicate(lambda ch: ch == '+'))

    def testUngetch(self):
        b = self.bufferFromString("ong")
        b.ungetch('y')
        self.assertEqual('y', b.getch())
        self.assertEqual('o', b.getch())
        b.ungetch('u')
        self.assertEqual('u', b.getch())
        self.assertEqual('n', b.getch())
        self.assertEqual('g', b.getch())
        self.assertEqual('', b.getch())

    def testRepeatedGetchOnEmptyStreamIsOk(self):
        b = self.bufferFromString("")
        self.assertEqual('', b.getch())
        self.assertEqual('', b.getch())

    def testCappedBuffer(self):
        b = CappedBuffer(self.bufferFromString("supercalifragilisticexpialidocious"), 5)
        self.assertEqual("s", b.getch())
        self.assertEqual("u", b.getch())
        self.assertEqual("p", b.getch())
        self.assertEqual("e", b.getch())
        self.assertEqual("r", b.getch())
        self.assertEqual('', b.getch())
        self.assertEqual('', b.getch())
        b.ungetch('r')
        self.assertEqual("r", b.getch())
        self.assertEqual('', b.getch())

    def testWhitespaceScanning(self):
        b = self.bufferFromString("    42\n43")
        self.assertEqual("    ", handleWhitespace(b))
        self.assertEqual("", handleWhitespace(b))
        self.assertEqual("4", b.getch())

    def testDecimalDigitScanning(self):
        b = self.bufferFromString("42 43!44")
        self.assertEqual(42, handleDecimalInt(b))
        self.assertEqual(" ", handleWhitespace(b))
        self.assertEqual(43, handleDecimalInt(b))

        b2 = self.bufferFromString("-1-2+3-4")
        self.assertEqual(-1, handleDecimalInt(b2))
        self.assertEqual(-2, handleDecimalInt(b2))
        self.assertEqual(3, handleDecimalInt(b2))
        self.assertEqual(-4, handleDecimalInt(b2))
        self.assertRaises(FormatError, handleDecimalInt, b2)

    def testCharacter(self):
        b = self.bufferFromString("hi!")
        self.assertEqual("h", handleChar(b))
        self.assertEqual("i", handleChar(b))
        self.assertEqual("!", handleChar(b))
        self.assertRaises(FormatError, handleChar, b)

    def testString(self):
        b = self.bufferFromString("-42 + 1 equals -41")
        self.assertEqual("-42", handleString(b))
        handleWhitespace(b)
        self.assertEqual("+", handleString(b))
        handleWhitespace(b)
        self.assertEqual("1", handleString(b))
        handleWhitespace(b)
        self.assertEqual("equals", handleString(b))
        handleWhitespace(b)
        self.assertEqual("-41", handleString(b))

    def testIntegerScanning(self):
        self.assertEqual((42, 43),
                          sscanf("   42\n   43  ", "%d %d"))
        self.assertEqual((8,), sscanf("10", "%o"))
        self.assertEqual((8,), sscanf("010", "%o"))
        self.assertEqual((15,), sscanf("F", "%x"))
        self.assertEqual((15,), sscanf("f", "%x"))
        self.assertEqual((15,), sscanf("0xF", "%x"))
        self.assertEqual((15,), sscanf("0XF", "%x"))
        self.assertEqual((15,), sscanf("0Xf", "%x"))
        self.assertEqual((-1, -2, 3, -4), sscanf("-1-2+3-4", "%d%d%d%d"))

    def testWordScanning(self):
        self.assertEqual(("hello", "world"),
                          sscanf("   hello world", "%s %s"))

    def testSuppression(self):
        self.assertEqual((), sscanf(" hello world", "%*s %*s"))
        self.assertEqual(("happy",),
                          sscanf("hello happy world", "%*s %s %*s"))
        self.assertEqual((), sscanf("h", "%*c"))

    def testWidth(self):
        self.assertEqual(("00010",), sscanf("00010101010111", "%5c"))
        self.assertEqual(("xy",), sscanf("xyz", "%2s"))
        self.assertEqual(("xy",), sscanf("              xyz", "%2s"))
        self.assertEqual(("  ",), sscanf("              xyz", "%2c"))

    def testFscanf(self):
        b = StringIO("hello world")
        self.assertEqual(("hello", " ", "world"), fscanf(b, "%s%c%s"))
        # Check that calling fscanf() twice doesn't
        # drop the last character
        b2 = StringIO("hello world")
        self.assertEqual(("hello",), fscanf(b2, "%s"))
        self.assertEqual((" ",), fscanf(b2, "%c"))
        self.assertEqual(("world",), fscanf(b2, "%s"))

    def testSkipLeadingSpaceOnScanning(self):
        """Ralph Heinkel reported a bug where floats weren't being
        parsed properly if there were leading whitespace for %f.
        This case checks that"""
        self.assertEqual((42.0,),
                          sscanf("    42.0", "%f"))

    def testFloats(self):
        self.assertEqual((3.14,
                           10.,
                           .001,
                           1e100,
                           3.14e-10,
                           0e0,), sscanf("""3.14
                           10.
                           .001
                           1e100
                           3.14e-10
                           0e0""", "%f %f %f %f %f %f"))

    def testMoreSimpleScanningExamples(self):
        self.assertEqual((192,168,1,1),
                          sscanf("192.168.1.1", "%d.%d.%d.%d"))
        self.assertEqual(("a", "b", "c"),
                          sscanf("  ab   c  ", "%1s%1s%s"))
        self.assertEqual(("hello", " ", "world"),
                          sscanf("hello world", "%s%c%s"))
        self.assertRaises(IncompleteCaptureError,
                          sscanf, "192.168.1.1", "%d %d %d %d")
        self.assertEqual(("danny",),
                          sscanf("hi danny", "hi %s"))
        self.assertEqual(("danny",),
                          sscanf("  hi danny", "  hi %s"))
        self.assertEqual(("a", "b", 3),
                          sscanf("ab3", "%c%c%d"))
        # this case is weird, but it happens in C too!
        self.assertRaises(IncompleteCaptureError,
                          sscanf, "  hi danny", "hi %s")

        # The example that's used in http://docs.python.org/lib/node109.html
        self.assertEqual(("/usr/bin/sendmail", 0, 4),
                          sscanf("/usr/bin/sendmail - 0 errors, 4 warnings",
                                 "%s - %d errors, %d warnings"))

    def testErroneousFormats(self):
        self.assertRaises(FormatError, compile, "%")
        self.assertRaises(FormatError, compile, "% ")
        self.assertRaises(FormatError, compile, "%*")
        self.assertRaises(FormatError, compile, "%*z")
        self.assertRaises(FormatError, compile, "% d")
        self.assertRaises(FormatError, compile, "%* d")


if __name__ == '__main__':
    unittest.main()



from typing import Iterable

from .typing_utils import RdPoint


def cross(o: RdPoint, a: RdPoint, b: RdPoint) -> float:
    """
    2D cross product of OA and OB vectors, i.e. z-component of their 3D cross product.
    Returns a positive value, if OAB makes a counter-clockwise turn,
    negative for clockwise turn, and zero if the points are collinear.
    """
    return (a.rate - o.rate) * (b.metric - o.metric) - (a.metric - o.metric) * (b.rate - o.rate)


def calculate_convex_hull(points: Iterable[RdPoint]) -> list[RdPoint]:
    """
    Computes the convex hull of a set of RdPoints.

    Input: an iterable sequence of RdPoints.
    Output: a list of RdPoints of the convex hull in counter-clockwise order,
    starting from the vertex with the lexicographically smallest coordinates.

    Implements Andrew's monotone chain algorithm. O(n log n) complexity.
    For this application only the lower hull is used, so upper hull is not calculated.
    Points that are nonmonotonic (higher rate, lower quality than some previous point) are disregarded.
    """

    # Sort the points lexicographically (tuples are compared lexicographically).
    # Remove duplicates to detect the case we have just one unique point.
    points = sorted(set(points), key=lambda point: (point.rate, point.metric))

    # Remove points that are nonmonotonic: higher rate, lower quality
    monotonic_points: list[RdPoint] = []
    for point in points:
        if len(monotonic_points) == 0 or point.metric >= monotonic_points[-1].metric:
            monotonic_points.append(point)

    # Boring case: no points or a single point, possibly repeated multiple times.
    if len(monotonic_points) == 0:
        return monotonic_points
    if len(monotonic_points) == 1:
        return monotonic_points

    # Build lower hull
    hull: list[RdPoint] = []
    for p in reversed(monotonic_points):
        while len(hull) >= 2 and cross(hull[-2], hull[-1], p) <= 0:
            hull.pop()
        hull.append(p)

    return hull



from vmaf import ExternalProgramCaller
from vmaf.core.feature_extractor import VmafexecFeatureExtractorMixin, FeatureExtractor


class FloatMotionFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "float_motion_feature"
    # VERSION = "1.0"
    VERSION = "1.1"  # add debug features

    ATOM_FEATURES = ['motion2',
                     'motion',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'motion2': 'motion2',
        'motion': 'motion',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_motion', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict, **optional_dict2})


class IntegerMotionFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "integer_motion_feature"
    # VERSION = "1.0"
    # VERSION = "1.1"  # vectorization
    VERSION = "1.2"  # add debug features

    ATOM_FEATURES = ['motion2',
                     'motion',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'motion2': 'integer_motion2',
        'motion': 'integer_motion',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'motion', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger,
            options={**optional_dict, **optional_dict2})


class FloatVifFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "float_VIF_feature"
    # VERSION = "1.0"
    VERSION = "1.1"  # add debug features

    ATOM_FEATURES = [
                     'vif_scale0', 'vif_scale1', 'vif_scale2', 'vif_scale3',
                     'vif', 'vif_num', 'vif_den',
                     'vif_num_scale0',
                     'vif_den_scale0',
                     'vif_num_scale1',
                     'vif_den_scale1',
                     'vif_num_scale2',
                     'vif_den_scale2',
                     'vif_num_scale3',
                     'vif_den_scale3',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'vif_scale0': 'vif_scale0',
        'vif_scale1': 'vif_scale1',
        'vif_scale2': 'vif_scale2',
        'vif_scale3': 'vif_scale3',
        'vif': 'vif',
        'vif_num': 'vif_num',
        'vif_den': 'vif_den',
        'vif_num_scale0': 'vif_num_scale0',
        'vif_den_scale0': 'vif_den_scale0',
        'vif_num_scale1': 'vif_num_scale1',
        'vif_den_scale1': 'vif_den_scale1',
        'vif_num_scale2': 'vif_num_scale2',
        'vif_den_scale2': 'vif_den_scale2',
        'vif_num_scale3': 'vif_num_scale3',
        'vif_den_scale3': 'vif_den_scale3',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_vif', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict, **optional_dict2})


class IntegerVifFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "integer_VIF_feature"
    # VERSION = "1.0"
    # VERSION = "1.1b"  # vif_enhn_gain_limit with matching_matlab code
    # VERSION = "1.1c"  # update boundary calculation
    # VERSION = "1.1d"  # update to use log2f to replace log2f_approx
    # VERSION = "1.2"  # fix vectorization corner cases
    VERSION = "1.3"  # add debug features

    ATOM_FEATURES = [
                     'vif_scale0', 'vif_scale1', 'vif_scale2', 'vif_scale3',
                     'vif', 'vif_num', 'vif_den',
                     'vif_num_scale0',
                     'vif_den_scale0',
                     'vif_num_scale1',
                     'vif_den_scale1',
                     'vif_num_scale2',
                     'vif_den_scale2',
                     'vif_num_scale3',
                     'vif_den_scale3',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'vif_scale0': 'integer_vif_scale0',
        'vif_scale1': 'integer_vif_scale1',
        'vif_scale2': 'integer_vif_scale2',
        'vif_scale3': 'integer_vif_scale3',
        'vif': 'integer_vif',
        'vif_num': 'integer_vif_num',
        'vif_den': 'integer_vif_den',
        'vif_num_scale0': 'integer_vif_num_scale0',
        'vif_den_scale0': 'integer_vif_den_scale0',
        'vif_num_scale1': 'integer_vif_num_scale1',
        'vif_den_scale1': 'integer_vif_den_scale1',
        'vif_num_scale2': 'integer_vif_num_scale2',
        'vif_den_scale2': 'integer_vif_den_scale2',
        'vif_num_scale3': 'integer_vif_num_scale3',
        'vif_den_scale3': 'integer_vif_den_scale3',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'vif', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict, **optional_dict2})


class FloatAdmFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "float_ADM_feature"
    # VERSION = "1.0"
    VERSION = "1.1"  # add debug features

    ATOM_FEATURES = ['adm2',
                     'adm_scale0',
                     'adm_scale1',
                     'adm_scale2',
                     'adm_scale3',
                     'adm',
                     'adm_num',
                     'adm_den',
                     'adm_num_scale0',
                     'adm_den_scale0',
                     'adm_num_scale1',
                     'adm_den_scale1',
                     'adm_num_scale2',
                     'adm_den_scale2',
                     'adm_num_scale3',
                     'adm_den_scale3',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'adm2': 'adm2',
        'adm_scale0': 'adm_scale0',
        'adm_scale1': 'adm_scale1',
        'adm_scale2': 'adm_scale2',
        'adm_scale3': 'adm_scale3',
        'adm': 'adm',
        'adm_num': 'adm_num',
        'adm_den': 'adm_den',
        'adm_num_scale0': 'adm_num_scale0',
        'adm_den_scale0': 'adm_den_scale0',
        'adm_num_scale1': 'adm_num_scale1',
        'adm_den_scale1': 'adm_den_scale1',
        'adm_num_scale2': 'adm_num_scale2',
        'adm_den_scale2': 'adm_den_scale2',
        'adm_num_scale3': 'adm_num_scale3',
        'adm_den_scale3': 'adm_den_scale3',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_adm', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict, **optional_dict2})


class IntegerPsnrFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = 'integer_PSNR_feature'
    VERSION = "1.0"

    ATOM_FEATURES = ['psnr_y', 'psnr_cb', 'psnr_cr']

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'psnr_y': 'psnr_y',
        'psnr_cb': 'psnr_cb',
        'psnr_cr': 'psnr_cr',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'psnr', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict, **optional_dict2})


class IntegerAdmFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "integer_ADM_feature"
    # VERSION = "1.0"
    # VERSION = "1.1"  # vectorization; small numerical diff introduced by adm_enhn_gain_limit
    VERSION = "1.2"  # add debug features

    ATOM_FEATURES = ['adm2',
                     'adm_scale0',
                     'adm_scale1',
                     'adm_scale2',
                     'adm_scale3',
                     'adm',
                     'adm_num',
                     'adm_den',
                     'adm_num_scale0',
                     'adm_den_scale0',
                     'adm_num_scale1',
                     'adm_den_scale1',
                     'adm_num_scale2',
                     'adm_den_scale2',
                     'adm_num_scale3',
                     'adm_den_scale3',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'adm2': 'integer_adm2',
        'adm_scale0': 'integer_adm_scale0',
        'adm_scale1': 'integer_adm_scale1',
        'adm_scale2': 'integer_adm_scale2',
        'adm_scale3': 'integer_adm_scale3',
        'adm': 'integer_adm',
        'adm_num': 'integer_adm_num',
        'adm_den': 'integer_adm_den',
        'adm_num_scale0': 'integer_adm_num_scale0',
        'adm_den_scale0': 'integer_adm_den_scale0',
        'adm_num_scale1': 'integer_adm_num_scale1',
        'adm_den_scale1': 'integer_adm_den_scale1',
        'adm_num_scale2': 'integer_adm_num_scale2',
        'adm_den_scale2': 'integer_adm_den_scale2',
        'adm_num_scale3': 'integer_adm_num_scale3',
        'adm_den_scale3': 'integer_adm_den_scale3',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'adm', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict, **optional_dict2})


class CIEDE2000FeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = 'CIEDE2000_feature'
    VERSION = "1.0"

    ATOM_FEATURES = ['ciede2000']

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'ciede2000': 'ciede2000',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type=self._get_workfile_yuv_type(asset)
        ref_path=asset.ref_procfile_path
        dis_path=asset.dis_procfile_path
        w=quality_width
        h=quality_height
        logger = self.logger

        optional_dict = self.optional_dict if self.optional_dict is not None else dict()
        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'ciede', yuv_type, ref_path, dis_path, w, h, log_file_path, logger,
            options={**optional_dict, **optional_dict2})



import unittest

import numpy as np

from vmaf.config import VmafConfig
from vmaf.core.adm_dwt2_cy import adm_dwt2_cy
from vmaf.core.adm_dwt2_py import adm_dwt2_py, adm_idwt2_py
from vmaf.tools.reader import YuvReader


class AdmDwt2PyTest(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def test_adm_dwt2_py(self):
        np.random.seed(0)
        x = np.random.uniform(low=-128, high=127, size=[324, 576]).astype(np.float64)
        a, v, h, d = adm_dwt2_py(x)
        self.assertEqual(a.shape, (162, 288))
        self.assertEqual(v.shape, (162, 288))
        self.assertEqual(h.shape, (162, 288))
        self.assertEqual(d.shape, (162, 288))
        self.assertAlmostEqual(float(np.std(a)), 73.8959273922819, places=5)
        self.assertAlmostEqual(float(np.std(v)), 73.69196319580078, places=5)
        self.assertAlmostEqual(float(np.std(h)), 73.53559112548828, places=5)
        self.assertAlmostEqual(float(np.std(d)), 73.52173614501953, places=5)

    def test_adm_dwt2_idwt2_py(self):
        np.random.seed(0)
        x = np.random.uniform(low=-128, high=127, size=[324, 576]).astype(np.float64)
        a, v, h, d = adm_dwt2_py(x)
        x2 = adm_idwt2_py((a, v, h, d))
        self.assertAlmostEqual(np.abs(np.max(x - x2)), 0.0, places=10)


class AdmDwt2CyTest(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def test_adm_dwt2_cy(self):
        np.random.seed(0)
        x = np.random.uniform(low=-128, high=127, size=[324, 576]).astype(np.float64)
        a, v, h, d = adm_dwt2_cy(x)
        self.assertEqual(a.shape, (162, 288))
        self.assertEqual(v.shape, (162, 288))
        self.assertEqual(h.shape, (162, 288))
        self.assertEqual(d.shape, (162, 288))
        self.assertAlmostEqual(float(np.std(a)), 73.94278958581368, places=5)
        self.assertAlmostEqual(float(np.std(v)), 73.61917114257812, places=5)
        self.assertAlmostEqual(float(np.std(h)), 73.30349892103438, places=5)
        self.assertAlmostEqual(float(np.std(d)), 73.19024658203125, places=5)

    def test_adm_dwt2_cy_small(self):
        np.random.seed(0)
        x = np.random.uniform(low=-128, high=127, size=[36, 44]).astype(np.float64)
        a, v, h, d = adm_dwt2_cy(x)
        self.assertEqual(a.shape, (18, 22))
        self.assertEqual(v.shape, (18, 22))
        self.assertEqual(h.shape, (18, 22))
        self.assertEqual(d.shape, (18, 22))
        self.assertAlmostEqual(float(np.std(a)), 75.67939583965345, places=5)
        self.assertAlmostEqual(float(np.std(v)), 74.84364345847594, places=5)
        self.assertAlmostEqual(float(np.std(h)), 73.1595099161633, places=5)
        self.assertAlmostEqual(float(np.std(d)), 68.17639102914511, places=5)

    def test_adm_dwt2_cy_xsmall(self):
        np.random.seed(0)
        x = np.random.uniform(low=-128, high=127, size=[18, 22]).astype(np.float64)
        a, v, h, d = adm_dwt2_cy(x)
        self.assertEqual(a.shape, (9, 11))
        self.assertEqual(v.shape, (9, 11))
        self.assertEqual(h.shape, (9, 11))
        self.assertEqual(d.shape, (9, 11))
        self.assertAlmostEqual(float(np.std(a)), 68.79495062113115, places=5)
        self.assertAlmostEqual(float(np.std(v)), 67.96417675879844, places=5)
        self.assertAlmostEqual(float(np.std(h)), 75.05987575325626, places=5)
        self.assertAlmostEqual(float(np.std(d)), 62.93524366183755, places=5)

    def test_adm_dwt2_cy_xsmallP_dc(self):
        x = (55 * np.ones([18, 22])).astype(np.float64)
        a, v, h, d = adm_dwt2_cy(x)

        self.assertEqual(a.shape, (9, 11))
        self.assertEqual(v.shape, (9, 11))
        self.assertEqual(h.shape, (9, 11))
        self.assertEqual(d.shape, (9, 11))

        self.assertAlmostEqual(float(np.max(a)), 110.0, places=12)
        self.assertAlmostEqual(float(np.max(v)), 0.0, places=12)
        self.assertAlmostEqual(float(np.max(h)), 0.0, places=12)
        self.assertAlmostEqual(float(np.max(d)), 0.0, places=12)

        self.assertAlmostEqual(float(np.min(a)), 110.0, places=12)
        self.assertAlmostEqual(float(np.min(v)), 0.0, places=12)
        self.assertAlmostEqual(float(np.min(h)), 0.0, places=12)
        self.assertAlmostEqual(float(np.min(d)), 0.0, places=12)

    def test_adm_dwt2_cy_dc(self):
        x = (55 * np.ones([324, 576])).astype(np.float64)
        a, v, h, d = adm_dwt2_cy(x)

        self.assertEqual(a.shape, (162, 288))
        self.assertEqual(v.shape, (162, 288))
        self.assertEqual(h.shape, (162, 288))
        self.assertEqual(d.shape, (162, 288))

        self.assertAlmostEqual(float(np.max(a)), 110.0, places=12)
        self.assertAlmostEqual(float(np.max(v)), 0.0, places=12)
        self.assertAlmostEqual(float(np.max(h)), 0.0, places=12)
        self.assertAlmostEqual(float(np.max(d)), 0.0, places=12)

        self.assertAlmostEqual(float(np.min(a)), 110.0, places=12)
        self.assertAlmostEqual(float(np.min(v)), 0.0, places=12)
        self.assertAlmostEqual(float(np.min(h)), 0.0, places=12)
        self.assertAlmostEqual(float(np.min(d)), 0.0, places=12)


class AdmDwt2CyTestOnAkiyo(unittest.TestCase):

    def setUp(self) -> None:
        with YuvReader(
                filepath=VmafConfig.test_resource_path("yuv", "refp_vmaf_hacking_investigation_0_0_akiyo_cif_notyuv_0to0_identity_vs_akiyo_cif_notyuv_0to0_multiply_q_352x288"),
                width=352, height=288, yuv_type='yuv420p'
        ) as yuv_reader_ref:
            self.y_ref = yuv_reader_ref.next()[0].astype(np.float64)
        with YuvReader(
                filepath=VmafConfig.test_resource_path("yuv", "disp_vmaf_hacking_investigation_0_0_akiyo_cif_notyuv_0to0_identity_vs_akiyo_cif_notyuv_0to0_multiply_q_352x288"),
                width=352, height=288, yuv_type='yuv420p'
        ) as yuv_reader_ref:
            self.y_dis = yuv_reader_ref.next()[0].astype(np.float64)

    def test_adm_dwt2_cy_on_akiyo_single_scale(self):

        a, v, h, d = adm_dwt2_cy(self.y_ref - 128.0)
        self.assertEqual(a.shape, (144, 176))
        self.assertEqual(v.shape, (144, 176))
        self.assertEqual(h.shape, (144, 176))
        self.assertEqual(d.shape, (144, 176))
        self.assertAlmostEqual(float(np.std(a)), 94.61946432328955, places=5)
        self.assertAlmostEqual(float(np.std(v)), 8.238639778464304, places=5)
        self.assertAlmostEqual(float(np.std(h)), 4.652637049444403, places=5)
        self.assertAlmostEqual(float(np.std(d)), 2.1727321628143614, places=5)
        self.assertAlmostEqual(float(np.mean(a)), -68.27941073300276, places=5)
        self.assertAlmostEqual(float(np.mean(v)), -0.05449070177725589, places=5)
        self.assertAlmostEqual(float(np.mean(h)), 0.004257626855952768, places=5)
        self.assertAlmostEqual(float(np.mean(d)), -0.002311283312114316, places=5)

        a, v, h, d = adm_dwt2_cy(self.y_dis - 128.0)
        self.assertEqual(a.shape, (144, 176))
        self.assertEqual(v.shape, (144, 176))
        self.assertEqual(h.shape, (144, 176))
        self.assertEqual(d.shape, (144, 176))
        self.assertAlmostEqual(float(np.std(a)), 108.79343831916911, places=5)
        self.assertAlmostEqual(float(np.std(v)), 9.51322612796857, places=5)
        self.assertAlmostEqual(float(np.std(h)), 5.4237937163381, places=5)
        self.assertAlmostEqual(float(np.std(d)), 2.6714361298075886, places=5)
        self.assertAlmostEqual(float(np.mean(a)), -41.14129780482415, places=4)
        self.assertAlmostEqual(float(np.mean(v)), -0.05419786586456123, places=5)
        self.assertAlmostEqual(float(np.mean(h)), 0.002907897856746483, places=5)
        self.assertAlmostEqual(float(np.mean(d)), -0.000920162342243558, places=5)
        self.assertAlmostEqual(float(np.max(a)), 246.64342088710973, places=5)
        self.assertAlmostEqual(float(np.max(v)), 153.63339344423332, places=5)
        self.assertAlmostEqual(float(np.max(h)), 120.05215038354704, places=5)
        self.assertAlmostEqual(float(np.max(d)), 113.9025450046058, places=5)
        self.assertAlmostEqual(float(np.min(a)), -269.44402359826955, places=4)
        self.assertAlmostEqual(float(np.min(v)), -128.21006288062947, places=5)
        self.assertAlmostEqual(float(np.min(h)), -101.95207793605813, places=5)
        self.assertAlmostEqual(float(np.min(d)), -51.79502250236081, places=5)

    def test_adm_dwt2_cy_on_akiyo_n_scales(self):

        a = self.y_dis - 128.0

        a, v, h, d = adm_dwt2_cy(a)
        self.assertEqual(a.shape, (144, 176))
        self.assertEqual(v.shape, (144, 176))
        self.assertEqual(h.shape, (144, 176))
        self.assertEqual(d.shape, (144, 176))
        self.assertAlmostEqual(float(np.std(a)), 108.79343831916911, places=5)
        self.assertAlmostEqual(float(np.std(v)), 9.51322612796857, places=5)
        self.assertAlmostEqual(float(np.std(h)), 5.4237937163381, places=5)
        self.assertAlmostEqual(float(np.std(d)), 2.6714361298075886, places=5)

        a, v, h, d = adm_dwt2_cy(a)
        self.assertEqual(a.shape, (72, 88))
        self.assertEqual(v.shape, (72, 88))
        self.assertEqual(h.shape, (72, 88))
        self.assertEqual(d.shape, (72, 88))
        self.assertAlmostEqual(float(np.std(a)), 214.13138192032943, places=4)
        self.assertAlmostEqual(float(np.std(v)), 25.622761528555454, places=5)
        self.assertAlmostEqual(float(np.std(h)), 17.70038682132729, places=5)
        self.assertAlmostEqual(float(np.std(d)), 11.180386017176657, places=5)

        a, v, h, d = adm_dwt2_cy(a)
        self.assertEqual(a.shape, (36, 44))
        self.assertEqual(v.shape, (36, 44))
        self.assertEqual(h.shape, (36, 44))
        self.assertEqual(d.shape, (36, 44))
        self.assertAlmostEqual(float(np.std(a)), 417.73602985019386, places=5)
        self.assertAlmostEqual(float(np.std(v)), 56.12149501761274, places=5)
        self.assertAlmostEqual(float(np.std(h)), 45.419327754665, places=5)
        self.assertAlmostEqual(float(np.std(d)), 23.307968058054456, places=5)

        a, v, h, d = adm_dwt2_cy(a)
        self.assertEqual(a.shape, (18, 22))
        self.assertEqual(v.shape, (18, 22))
        self.assertEqual(h.shape, (18, 22))
        self.assertEqual(d.shape, (18, 22))
        self.assertAlmostEqual(float(np.std(a)), 800.3512673377832, places=5)
        self.assertAlmostEqual(float(np.std(v)), 133.36092909420165, places=5)
        self.assertAlmostEqual(float(np.std(h)), 113.7904328069727, places=5)
        self.assertAlmostEqual(float(np.std(d)), 47.651801810101496, places=5)

        a, v, h, d = adm_dwt2_cy(a)
        self.assertEqual(a.shape, (9, 11))
        self.assertEqual(v.shape, (9, 11))
        self.assertEqual(h.shape, (9, 11))
        self.assertEqual(d.shape, (9, 11))
        self.assertAlmostEqual(float(np.std(a)), 1477.3950051168213, places=5)
        self.assertAlmostEqual(float(np.std(v)), 284.3777998557365, places=5)
        self.assertAlmostEqual(float(np.std(h)), 324.15282224075446, places=5)
        self.assertAlmostEqual(float(np.std(d)), 116.24440187217311, places=5)


class AdmDwt2CyTestOnAkiyoXsmall(unittest.TestCase):

    def setUp(self) -> None:
        with YuvReader(
                filepath=VmafConfig.test_resource_path("yuv", "refp_vmaf_hacking_investigation_0_0_akiyo_cif_notyuv_0to0_identity_vs_akiyo_cif_notyuv_0to0_multiply_q_18x22"),
                width=22, height=18, yuv_type='yuv420p'
        ) as yuv_reader_ref:
            self.y_ref = yuv_reader_ref.next()[0].astype(np.float64)
        with YuvReader(
                filepath=VmafConfig.test_resource_path("yuv", "disp_vmaf_hacking_investigation_0_0_akiyo_cif_notyuv_0to0_identity_vs_akiyo_cif_notyuv_0to0_multiply_q_18x22"),
                width=22, height=18, yuv_type='yuv420p'
        ) as yuv_reader_ref:
            self.y_dis = yuv_reader_ref.next()[0].astype(np.float64)

    def test_adm_dwt2_cy_on_akiyo_single_scale(self):

        a, v, h, d = adm_dwt2_cy((self.y_ref - 128.0).astype(np.float64))
        self.assertEqual(a.shape, (9, 11))
        self.assertEqual(v.shape, (9, 11))
        self.assertEqual(h.shape, (9, 11))
        self.assertEqual(d.shape, (9, 11))
        self.assertAlmostEqual(float(np.std(a)), 70.96065160084203, places=5)
        self.assertAlmostEqual(float(np.std(v)), 14.048580053657007, places=5)
        self.assertAlmostEqual(float(np.std(h)), 44.79487471014486, places=5)
        self.assertAlmostEqual(float(np.std(d)), 17.35515294883506, places=5)
        self.assertAlmostEqual(float(np.mean(a)), -75.84326395274935, places=5)
        self.assertAlmostEqual(float(np.mean(v)), -3.2531880028079962, places=5)
        self.assertAlmostEqual(float(np.mean(h)), 1.818506248770275, places=5)
        self.assertAlmostEqual(float(np.mean(d)), -0.3250551358064137, places=5)

        self.assertAlmostEqual(a[0][0], -145.9108983848698, places=6)
        self.assertAlmostEqual(a[0][-1], -140.54543100196207, places=6)
        self.assertAlmostEqual(a[1][0], -136.4577676021147, places=6)
        self.assertAlmostEqual(a[1][-1], -120.36561690415635, places=6)
        self.assertAlmostEqual(a[-1][0], 79.6265877365092, places=5)
        self.assertAlmostEqual(a[-1][-1], 44.29716825973104, places=6)


if __name__ == '__main__':
    unittest.main(verbosity=2)



from __future__ import absolute_import

import os
import ssl
import urllib.request
import urllib.error

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

PYTHON_ROOT = os.path.dirname(os.path.realpath(__file__))
ROOT = os.path.abspath(os.path.join(PYTHON_ROOT, '..', '..',))
VMAF_RESOURCE_ROOT = "https://github.com/Netflix/vmaf_resource/raw/master"


def download_reactively(local_path, remote_path):
    if not os.path.exists(local_path):
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        print(f'download {local_path} from {remote_path}')
        try:
            ssl._create_default_https_context = ssl._create_unverified_context
            urllib.request.urlretrieve(remote_path, local_path)
        except urllib.error.HTTPError as e:
            print(f"error downloading from {remote_path}")
            raise e


class VmafExternalConfig(object):

    _MISSING_EXTERNAL_MESSAGE = """
    Must install {name} and set {key} in %s/externals.py, e.g. add a line like
    {key} = "[path to exec]/{name}"
    """ % PYTHON_ROOT

    @staticmethod
    def _path_from_external(name):
        """
        :param str name: Name of external configuration to look up
        :return str: Configured path, if any
        """
        try:
            from . import externals
            path = getattr(externals, name, None)
            if path and os.path.exists(path):
                return path
        except ImportError:
            print('ImportError')
            pass

        return None

    @staticmethod
    def _attr_from_external(name):
        """
        :param str name: Name of external configuration to look up
        :return str: Configured attribute (not path), if any
        """
        try:
            from . import externals
            attr = getattr(externals, name, None)
            return attr
        except ImportError:
            print('ImportError')
            pass
        return None

    @classmethod
    def ffmpeg_path(cls):
        """
        :return str: Path to ffmpeg, if installed and configured via `externals` module
        """
        return cls._path_from_external('FFMPEG_PATH')

    @classmethod
    def ffmpeg_env(cls):
        """
        :return dict: Dictionary of environmental variables when running ffmpeg, if installed and configured via
        'externals` module
        """
        return cls._attr_from_external('FFMPEG_ENV')

    @classmethod
    def matlab_path(cls):
        """
        :return str: Path to matlab, if installed and configured via `externals` module
        """
        return cls._path_from_external('MATLAB_PATH')

    @classmethod
    def matlab_runtime_path(cls):
        """
        :return str: Path to matlab runtime, if installed and configured via `externals` module
        """
        return cls._path_from_external('MATLAB_RUNTIME_PATH')

    @classmethod
    def cvx_path(cls):
        """
        :return str: Path to cvx, if installed and configured via `externals` module
        """
        return cls._path_from_external('CVX_PATH')

    @classmethod
    def psnr_path(cls):
        """
        :return str: Path to external psnr executable, if installed and configured via `externals` module
        """
        return cls._path_from_external('PSNR_PATH')

    @classmethod
    def moment_path(cls):
        """
        :return str: Path to external moment executable, if installed and configured via `externals` module
        """
        return cls._path_from_external('MOMENT_PATH')

    @classmethod
    def ssim_path(cls):
        """
        :return str: Path to external ssim executable, if installed and configured via `externals` module
        """
        return cls._path_from_external('SSIM_PATH')

    @classmethod
    def ms_ssim_path(cls):
        """
        :return str: Path to external ms_ssim executable, if installed and configured via `externals` module
        """
        return cls._path_from_external('MS_SSIM_PATH')

    @classmethod
    def vmaf_path(cls):
        """
        :return str: Path to external vmaf executable, if installed and configured via `externals` module
        """
        return cls._path_from_external('VMAF_PATH')

    @classmethod
    def vmafexec_path(cls):
        """
        :return str: Path to external vmafexec executable, if installed and configured via `externals` module
        """
        return cls._path_from_external('VMAFEXEC_PATH')

    @classmethod
    def get_and_assert_ffmpeg(cls):
        path = cls.ffmpeg_path()
        assert path is not None, cls._MISSING_EXTERNAL_MESSAGE.format(name='ffmpeg', key='FFMPEG_PATH')
        return path

    @classmethod
    def get_and_assert_matlab(cls):
        path = cls.matlab_path()
        assert path is not None, cls._MISSING_EXTERNAL_MESSAGE.format(name='matlab', key='MATLAB_PATH')
        return path

    @classmethod
    def get_and_assert_matlab_runtime(cls):
        path = cls.matlab_runtime_path()
        assert path is not None, \
            """Must install matlab runtime (v9.1) and set {key} in {root}/externals.py, e.g. add a line like {key} = "[path to matlab runtime]/v91"
            """.format(root=PYTHON_ROOT, key='MATLAB_RUNTIME_PATH')
        return path

    @classmethod
    def get_and_assert_cvx(cls):
        path = cls.cvx_path()
        assert path is not None, cls._MISSING_EXTERNAL_MESSAGE.format(name='cvx', key='CVX_PATH')
        return path


class VmafConfig(object):

    @classmethod
    def root_path(cls, *components):
        return os.path.join(ROOT, *components)

    @classmethod
    def file_result_store_path(cls, *components):
        return cls.root_path('workspace', 'result_store_dir', 'file_result_store', *components)

    @classmethod
    def encode_store_path(cls, *components):
        return cls.root_path('workspace', 'result_store_dir', 'encode_store', *components)

    @classmethod
    def workspace_path(cls, *components):
        return cls.root_path('workspace', *components)

    @classmethod
    def workdir_path(cls, *components):
        return cls.root_path('workspace', 'workdir', *components)

    @classmethod
    def model_path(cls, *components):
        return cls.root_path('model', *components)

    @classmethod
    def resource_path(cls, *components):
        return cls.root_path('resource', *components)

    @classmethod
    def test_resource_path(cls, *components, bypass_download=False):
        local_path = cls.root_path('python', 'test', 'resource', *components)
        if bypass_download:
            pass
        else:
            remote_path = os.path.join(VMAF_RESOURCE_ROOT, 'python', 'test', 'resource', *components)
            download_reactively(local_path, remote_path)
        return local_path

    @classmethod
    def tools_resource_path(cls, *components):
        return cls.root_path('python', 'vmaf', 'tools', 'resource', *components)

    @classmethod
    def encode_path(cls, *components):
        return cls.root_path('workspace', 'encode', *components)


class DisplayConfig(object):

    @staticmethod
    def show(**kwargs):
        from vmaf import plt
        if 'write_to_dir' in kwargs:
            fmt = kwargs['format'] if 'format' in kwargs else 'png'
            filedir = kwargs['write_to_dir'] if kwargs['write_to_dir'] is not None else VmafConfig.workspace_path('output')
            os.makedirs(filedir, exist_ok=True)
            for fignum in plt.get_fignums():
                fig = plt.figure(fignum)
                fig.savefig(os.path.join(filedir, str(fignum) + '.' + fmt), format=fmt)
        else:
            plt.show()



__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

from vmaf.core.feature_extractor import FeatureExtractor
from vmaf.core.result import BasicResult


class FeatureAssembler(object):
    """
    Assembles features for an input list of Assets on an input list of
    FeatureExtractors. For each asset, it outputs a BasicResult object.
    """

    def __init__(self,
                 feature_dict,
                 feature_option_dict,
                 assets,
                 logger,
                 fifo_mode,
                 delete_workdir,
                 result_store,
                 optional_dict=None,
                 optional_dict2=None,
                 parallelize=False,
                 processes=None,
                 save_workfiles=False,
                 ):
        """
        :param feature_dict: in the format of:
        {FeatureExtractor_type:'all', ...}, or
        {FeatureExtractor_type:[atom_features,], ...}.
        For example, the below are valid feature dicts:
        {'VMAF_feature':'all', 'BRISQUE_feature':'all'},
        {'VMAF_feature':['vif', 'ansnr'], 'BRISQUE_feature':'all'}
        :param feature_option_dict: contains options to extract a particular
        feature, for example:
        {'VMAF_feature':{'force_extraction':True}, 'BRISQUE_feature':{}},
        :param assets:
        :param logger:
        :param fifo_mode:
        :param delete_workdir:
        :param result_store:
        :param optional_dict:
        :param parallelize:
        :return:
        """
        self.feature_dict = feature_dict
        self.feature_option_dict = feature_option_dict
        self.assets = assets
        self.logger = logger
        self.fifo_mode = fifo_mode
        self.delete_workdir = delete_workdir
        self.result_store = result_store
        self.optional_dict = optional_dict
        self.optional_dict2 = optional_dict2
        self.parallelize = parallelize
        self.processes = processes
        self.save_workfiles = save_workfiles

        self.type2results_dict = {}

    def run(self):
        """
        Do all the calculation here.
        :return:
        """

        # for each FeatureExtractor_type key in feature_dict, find the subclass
        # of FeatureExtractor, run, and put results in a dict
        for fextractor_type in self.feature_dict:
            runner = self._get_fextractor_instance(fextractor_type)
            runner.run(parallelize=self.parallelize, processes=self.processes)
            results = runner.results
            self.type2results_dict[fextractor_type] = results

        result_dicts = self._create_feature_result_dicts()

        self.results = list(map(
            lambda tasset: BasicResult(tasset[0], tasset[1]),
            zip(self.assets, result_dicts)
        ))

    def _create_feature_result_dicts(self):
        # assemble an output dict with demanded atom features
        # atom_features_dict = self.fextractor_atom_features_dict
        result_dicts = list(map(lambda x: dict(), self.assets))
        for fextractor_type in self.feature_dict:
            assert fextractor_type in self.type2results_dict
            for atom_feature in self._get_atom_features(fextractor_type):
                scores_key = self._get_scores_key(fextractor_type, atom_feature)
                for result_index, result in enumerate(self.type2results_dict[fextractor_type]):
                    try:
                        result_dicts[result_index][scores_key] = result[scores_key]
                    except KeyError:
                        # Determine scores keys from other features. These need to be discarded when wildcard-querying
                        # for all scores keys pertinent to a particular atom feature.
                        other_scores_keys = [self._get_scores_key(fextractor_type, other_atom_feature)
                                             for other_atom_feature in self._get_atom_features(fextractor_type)
                                             if other_atom_feature is not atom_feature]
                        scores_key_alt = BasicResult.scores_key_wildcard_match(result.result_dict, scores_key,
                                                                               excluded_scores_keys=other_scores_keys)
                        result_dicts[result_index][scores_key] = result[scores_key_alt]
        return result_dicts

    def remove_results(self):
        """
        Remove all relevant Results stored in ResultStore, which is specified
        at the constructor.
        :return:
        """
        for fextractor_type in self.feature_dict:
            fextractor = self._get_fextractor_instance(fextractor_type)
            fextractor.remove_results()

    def _get_scores_key(self, fextractor_type, atom_feature):
        fextractor_subclass = FeatureExtractor.find_subclass(fextractor_type)
        scores_key = fextractor_subclass.get_scores_key(atom_feature)
        return scores_key

    def _get_atom_features(self, fextractor_type):
        if self.feature_dict[fextractor_type] == 'all':
            fextractor_class = FeatureExtractor.find_subclass(fextractor_type)
            atom_features = fextractor_class.ATOM_FEATURES + getattr(fextractor_class, 'DERIVED_ATOM_FEATURES', [])

        else:
            atom_features = self.feature_dict[fextractor_type]

        return atom_features

    def _get_fextractor_instance(self, fextractor_type):
        """
        On the assignment of feature extractor's optional_dict, the following rules are used:
        1) If input feature_option_dict has been assigned (for example, this could come from a "feature_opts_dicts"
        field in a model file passed to VmafQualityRunner or its subclass), use it.
        2) If input feature_option_dict is not assigned (i.e. None), then use the input optional_dict. This happens,
        for example, in VmafQualityRunner or its subclass, the model file does not specify a
        "feature_opts_dicts" field, but we pass the fields in through input optional_dict.
        """

        fextractor_class = FeatureExtractor.find_subclass(fextractor_type)

        if self.feature_option_dict is not None and fextractor_type in self.feature_option_dict:
            optional_dict = self.feature_option_dict[fextractor_type]
        else:
            optional_dict = self.optional_dict

        fextractor = fextractor_class(assets=self.assets,
                                      logger=self.logger,
                                      fifo_mode=self.fifo_mode,
                                      delete_workdir=self.delete_workdir,
                                      result_store=self.result_store,
                                      optional_dict=optional_dict,
                                      optional_dict2=self.optional_dict2,
                                      save_workfiles=self.save_workfiles,
                                      )
        return fextractor



import numpy as np

from vmaf.tools.interpolation_utils import InterpolationUtils

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class BDrateCalculator(object):
    """
    BD rate calculator. Implementation validated against JCTVC-E137:
    http://phenix.it-sudparis.eu/jct/doc_end_user/documents/5_Geneva/wg11/JCTVC-E137-v1.zip
    """

    REJECTED_BD_RATE_NO_OVERLAP = -10000
    REJECTED_BD_RATE_NON_MONOTONIC = -20000
    REJECTED_BD_RATE_MESSED_UP_RATE = -30000
    REJECTED_BALANCE_BAD_MEASUREMENTS = -40000

    @staticmethod
    def _dedup_and_order(set_):
        return sorted(list(set(set_)), key=lambda x: x[0])

    @classmethod
    def CalcBDRate(cls, setA, setB):

        # ==== added by zli =======
        setA = cls._dedup_and_order(setA)
        setB = cls._dedup_and_order(setB)
        # ==== added by zli =======

        assert not (len(setA) < 4 or len(setB) < 4), \
            "Problem with input RD point lists. setA is size {}, " \
            "setB is size {}".format(len(setA), len(setB))

        if not cls.isCurveMonotonic(setA):
            raise AssertionError(cls.REJECTED_BD_RATE_NON_MONOTONIC)

        if not cls.isCurveMonotonic(setB):
            raise AssertionError(cls.REJECTED_BD_RATE_NON_MONOTONIC)

        if not cls.ratesLookOkay(setA):
            raise AssertionError(cls.REJECTED_BD_RATE_MESSED_UP_RATE)

        if not cls.ratesLookOkay(setB):
            raise AssertionError(cls.REJECTED_BD_RATE_MESSED_UP_RATE)

        minMainPSNR = setA[0][1]
        maxMainPSNR = setA[-1][1]
        minHighPSNR = setB[0][1]
        maxHighPSNR = setB[-1][1]

        minPSNR = max(minMainPSNR, minHighPSNR)
        maxPSNR = min(maxMainPSNR, maxHighPSNR)

        # no overlap, so mark it in a special way
        if minPSNR >= maxPSNR:
            return cls.REJECTED_BD_RATE_NO_OVERLAP

        vA = cls.bdrint(setA, minPSNR, maxPSNR)
        vB = cls.bdrint(setB, minPSNR, maxPSNR)

        avg = (vB - vA) / (maxPSNR - minPSNR)
        return np.power(10, avg) - 1

    @staticmethod
    def isCurveMonotonic(set_):
        for i in range(len(set_) - 1):

            # ==== added by zli =======
            if set_[i][0] >= set_[i + 1][0]:
                return False
            # ==== added by zli =======

            # ==== added by zli =======
            # if set_[i][1] > set_[i + 1][1]:
            if set_[i][1] >= set_[i + 1][1]:
            # ==== added by zli =======
                return False

        return True

    @staticmethod
    def ratesLookOkay(set_):
        for i in range(len(set_)):
            if set_[i][0] == 0:
                return False
        return True

    # // BD-rate calculation for arbitrary number (N) points
    # // cf. https://www.mathworks.com/moler/interp.pdf, sections 3.3 - 3.4
    @staticmethod
    def bdrint(rdPointsList, minPSNR, maxPSNR):

        N = len(rdPointsList)

        log_rate = []
        log_dist = []
        H = []
        delta = []
        d = []
        c = []
        b = []

        InterpolationUtils.computeParamsForSegments(rdPointsList, log_rate, log_dist, H, delta, d, c, b, True)

        # // cubic function is rate(i) + s*(d(i) + s*(c(i) + s*(b(i))) where s = x - dist(i)
        # // or rate(i) + s*d(i) + s*s*c(i) + s*s*s*b(i)
        # // primitive is s*rate(i) + s*s*d(i)/2 + s*s*s*c(i)/3 + s*s*s*s*b(i)/4

        result = 0.0

        for i in range(N - 1):
            s0 = log_dist[i]
            s1 = log_dist[i + 1]

            # // clip s0 to valid range
            s0 = max(s0, minPSNR)
            s0 = min(s0, maxPSNR)
            s0 -= log_dist[i]

            # // clip s1 to valid range
            s1 = max(s1, minPSNR)
            s1 = min(s1, maxPSNR)
            s1 -= log_dist[i]

            if s1 > s0:
                result += (s1 - s0) * log_rate[i]
                result += (s1 * s1 - s0 * s0) * d[i] / 2.0
                result += (s1 * s1 * s1 - s0 * s0 * s0) * c[i] / 3.0
                result += (s1 * s1 * s1 * s1 - s0 * s0 * s0 * s0) * b[i] / 4.0

        return result
